"""Base model classes
"""
from typing import Union, Dict

from aniemore.custom.modeling_classificators import (
    BaseModelForVoiceBaseClassification,
    BaseMultiModalForSequenceBaseClassification,
    AudioTextFusionModelForSequenceClassificaion,
    SpeechModelOutput,
    MultiModalConfig
)

import torch
from transformers import (
    WavLMForSequenceClassification,
    WavLMConfig,
    BertConfig,
    WavLMModel,
    BertModel,
    PretrainedConfig,
)

from transformers.models.wavlm.modeling_wavlm import (
    WavLMEncoder,
    WavLMEncoderStableLayerNorm,
    WavLMFeatureEncoder
)

from transformers.models.bert.modeling_bert import BertEncoder


class FusionConfig(MultiModalConfig):
    """Base class for fusion configs
    Just for fine-tuning models, no more

    Args:
        audio_config (PretrainedConfig): audio config
        text_config (PretrainedConfig): text config
        id2label (Dict[int, str]): id2label
        label2id (Dict[str, int]): label2id
        num_heads (int, optional): number of heads. Defaults to 8.
        kernel_size (int, optional): kernel size. Defaults to 1.
        pooling_mode (str, optional): pooling mode. Defaults to "mean".
        problem_type (str, optional): problem type. Defaults to "single_label_classification".
        gradient_checkpointing (bool, optional): gradient checkpointing. Defaults to True.
    """

    def __init__(
            self,
            audio_config: PretrainedConfig,
            text_config: PretrainedConfig,
            id2label: Dict[int, str],
            label2id: Dict[str, int],
            num_heads: int = 8,
            kernel_size: int = 1,
            pooling_mode: str = "mean",
            problem_type: str = "single_label_classification",
            gradient_checkpointing: bool = True,
            **kwargs):
        super().__init__(**kwargs)

        self.update({audio_config.architectures[0]: audio_config.to_dict()})
        self.update({text_config.architectures[0]: text_config.to_dict()})

        self.id2label = id2label
        self.label2id = label2id
        self.num_labels = len(id2label)
        self.num_heads = num_heads
        self.kernel_size = kernel_size
        self.pooling_mode = pooling_mode
        self.problem_type = problem_type
        self.gradient_checkpointing = gradient_checkpointing


class FusionModuleQ(torch.nn.Module):
    """FusionModuleQ is a fusion module for the query
    https://arxiv.org/abs/2302.13661
    https://arxiv.org/abs/2207.04697

    Args:
        audio_dim (int): audio dimension
        text_dim (int): text dimension
        num_heads (int): number of heads
    """

    def __init__(self, audio_dim, text_dim, num_heads):
        super().__init__()

        # pick the lowest dimension of the two modalities
        self.dimension = min(audio_dim, text_dim)

        # attention modules
        self.a_self_attention = torch.nn.MultiheadAttention(self.dimension, num_heads=num_heads)
        self.t_self_attention = torch.nn.MultiheadAttention(self.dimension, num_heads=num_heads)

        # layer norm
        self.audio_norm = torch.nn.LayerNorm(self.dimension)
        self.text_norm = torch.nn.LayerNorm(self.dimension)


class WavLMForVoiceClassification(BaseModelForVoiceBaseClassification):
    """WavLMForVoiceClassification is a model for voice classification task
     (e.g. speech command, voice activity detection, etc.)

    Args:
        config (WavLMConfig): config
        num_labels (int): number of labels

    Attributes:
        config (WavLMConfig): config
        num_labels (int): number of labels
        wavlm (WavLMForSequenceClassification): wavlm model
    """

    def __init__(self, config, num_labels):
        super().__init__(config, num_labels)
        self.wavlm = WavLMForSequenceClassification(config)
        self.init_weights()


class WavLMBertForSequenceClassification(BaseMultiModalForSequenceBaseClassification):
    """WavLMBertForSequenceClassification is a model for sequence classification task
     (e.g. sentiment analysis, text classification, etc.)

    Args:
        config (WavLMBertConfig): config

    Attributes:
        config (WavLMBertConfig): config
        audio_config (WavLMConfig): wavlm config
        text_config (BertConfig): bert config
        audio_model (WavLMModel): wavlm model
        text_model (BertModel): bert model
        classifier (torch.nn.Linear): classifier
    """

    def __init__(self, config, finetune=False):
        super().__init__(config)
        self.supports_gradient_checkpointing = getattr(config, "gradient_checkpointing", True)

        self.audio_config = WavLMConfig.from_dict(self.config.WavLMModel)
        self.text_config = BertConfig.from_dict(self.config.BertModel)

        if not finetune:
            self.audio_model = WavLMModel(self.audio_config)
            self.text_model = BertModel(self.text_config)

        else:
            self.audio_model = WavLMModel.from_pretrained(self.audio_config._name_or_path, config=self.audio_config)
            self.text_model = BertModel.from_pretrained(self.text_config._name_or_path, config=self.text_config)

        self.classifier = torch.nn.Linear(
            self.audio_config.hidden_size + self.text_config.hidden_size, self.num_labels
        )
        self.init_weights()

    @staticmethod
    def _set_gradient_checkpointing(module, value=False):
        if isinstance(module, (WavLMEncoder, WavLMEncoderStableLayerNorm, WavLMFeatureEncoder, BertEncoder)):
            module.gradient_checkpointing = value


class WavLMBertFusionForSequenceClassification(AudioTextFusionModelForSequenceClassificaion):
    """WavLMBertForSequenceClassification is a model for sequence classification task
     (e.g. sentiment analysis, text classification, etc.) for fine-tuning
    Args:
        config (WavLMBertConfig): config
    Attributes:
        config (WavLMBertConfig): config
        audio_config (WavLMConfig): wavlm config
        text_config (BertConfig): bert config
        audio_model (WavLMModel): wavlm model
        text_model (BertModel): bert model
        fusion_module_{i} (FusionModuleQ): Fusion Module Q
        audio_projector (Union[torch.nn.Linear, None]): Projection layer for audio embeds
        text_projector (Union[torch.nn.Linear, None]): Projection layer for text embeds
        audio_avg_pool (Union[torch.nn.AvgPool1d, None]): Audio average pool (out from fusion block)
        text_avg_pool (Union[torch.nn.AvgPool1d, None]): Text average pool (out from fusion block)
        classifier (torch.nn.Linear): classifier
    """

    def __init__(self, config, finetune=False):
        super().__init__(config)
        self.supports_gradient_checkpointing = getattr(config, "gradient_checkpointing", True)

        self.audio_config = WavLMConfig.from_dict(self.config.WavLMModel)
        self.text_config = BertConfig.from_dict(self.config.BertModel)

        if not finetune:
            self.audio_model = WavLMModel(self.audio_config)
            self.text_model = BertModel(self.text_config)

        else:
            self.audio_model = WavLMModel.from_pretrained(self.audio_config._name_or_path, config=self.audio_config)
            self.text_model = BertModel.from_pretrained(self.text_config._name_or_path, config=self.text_config)

        # fusion module with V3 strategy (one projection on entry, no projection in continuous)
        for i in range(self.config.num_fusion_layers):
            setattr(self, f"fusion_module_{i + 1}", FusionModuleQ(
                self.audio_config.hidden_size, self.text_config.hidden_size, self.config.num_heads
            ))

        self.audio_projector = torch.nn.Linear(self.audio_config.hidden_size, self.text_config.hidden_size)
        self.text_projector = torch.nn.Linear(self.text_config.hidden_size, self.text_config.hidden_size)

        # Avg Pool
        self.audio_avg_pool = torch.nn.AvgPool1d(self.config.kernel_size)
        self.text_avg_pool = torch.nn.AvgPool1d(self.config.kernel_size)

        # output dimensions of wav2vec2 and bert are 768 and 1024 respectively
        cls_dim = min(self.audio_config.hidden_size, self.text_config.hidden_size)
        self.classifier = torch.nn.Linear(
            (cls_dim * 2) // self.config.kernel_size, self.config.num_labels
        )

        self.init_weights()

    @staticmethod
    def _set_gradient_checkpointing(module, value=False):
        if isinstance(module, (WavLMEncoder, WavLMEncoderStableLayerNorm, WavLMFeatureEncoder, BertEncoder)):
            module.gradient_checkpointing = value

    def forward(
            self,
            input_ids=None,
            input_values=None,
            text_attention_mask=None,
            audio_attention_mask=None,
            token_type_ids=None,
            position_ids=None,
            head_mask=None,
            inputs_embeds=None,
            labels=None,
            output_attentions=None,
            output_hidden_states=None,
            return_dict=True,
    ):
        """Forward method for multimodal model for sequence classification task (e.g. text + audio)
        Args:
            input_ids (torch.LongTensor, optional): input ids. Defaults to None.
            input_values (torch.FloatTensor, optional): input values. Defaults to None.
            text_attention_mask (torch.LongTensor, optional): text attention mask. Defaults to None.
            audio_attention_mask (torch.LongTensor, optional): audio attention mask. Defaults to None.
            token_type_ids (torch.LongTensor, optional): token type ids. Defaults to None.
            position_ids (torch.LongTensor, optional): position ids. Defaults to None.
            head_mask (torch.FloatTensor, optional): head mask. Defaults to None.
            inputs_embeds (torch.FloatTensor, optional): inputs embeds. Defaults to None.
            labels (torch.LongTensor, optional): labels. Defaults to None.
            output_attentions (bool, optional): output attentions. Defaults to None.
            output_hidden_states (bool, optional): output hidden states. Defaults to None.
            return_dict (bool, optional): return dict. Defaults to True.
        Returns:
            torch.FloatTensor: logits
        """
        audio_output = self.audio_model(
            input_values=input_values,
            attention_mask=audio_attention_mask,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict
        )
        text_output = self.text_model(
            input_ids=input_ids,
            attention_mask=text_attention_mask,
            token_type_ids=token_type_ids,
            position_ids=position_ids,
            head_mask=head_mask,
            inputs_embeds=inputs_embeds,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        # Mean pooling
        audio_avg = self.merged_strategy(audio_output.last_hidden_state, mode=self.config.pooling_mode)

        # Projection
        audio_proj = self.audio_projector(audio_avg)
        text_proj = self.text_projector(text_output.pooler_output)

        audio_mha, text_mha = None, None

        for i in range(self.config.num_fusion_layers):
            fusion_module = getattr(self, f"fusion_module_{i + 1}")

            if i == 0:
                audio_mha, text_mha = fusion_module(audio_proj, text_proj)
            else:
                audio_mha, text_mha = fusion_module(audio_mha, text_mha)

        audio_avg = self.audio_avg_pool(audio_mha)
        text_avg = self.text_avg_pool(text_mha)

        fusion_output = torch.concat((audio_avg, text_avg), dim=1)

        logits = self.classifier(fusion_output)
        loss = None

        if labels is not None:
            loss = self.compute_loss(logits, labels)

        return SpeechModelOutput(
            loss=loss,
            logits=logits
        )



from typing import Union, List, Dict, Tuple
import numpy
import torch
import torchaudio

from aniemore.models import Model
from aniemore.utils.classes import (
    BaseRecognizer,
    RecognizerOutputOne,
    RecognizerOutputMany,
)

from transformers import PreTrainedTokenizerBase, BatchEncoding

from aniemore.utils.speech2text import Speech2Text


class VoiceTextRecognizer(BaseRecognizer):
    @classmethod
    def speech_file_to_array_fn(cls, path):
        """Загружаем аудиофайл в массив

        Args:
          path: путь к файлу

        Returns:
          numpy.ndarray

        """
        speech_array, _sampling_rate = torchaudio.load(path)
        resampler = torchaudio.transforms.Resample(_sampling_rate)
        speech = resampler(speech_array).squeeze().numpy()
        return speech

    @classmethod
    def tokenize(cls, text: Union[str, List[str]], tokenizer: PreTrainedTokenizerBase) -> BatchEncoding:
        """Токенизация текста

        Args:
          text: текст
          tokenizer: токенайзер

        Returns:
          список токенов

        """

        return tokenizer(
            text,
            padding="max_length",
            truncation=True,
            max_length=128,
            return_tensors="pt"
        )

    def _get_torch_scores(
            self,
            speech: Union[List[torch.Tensor], torch.Tensor, numpy.ndarray],
            text_inputs: Union[Dict[str, torch.Tensor], BatchEncoding]) -> torch.Tensor:
        """Получаем выход модели

        Args:
          speech: ndarray
          text_tokens: ndarray

        Returns:
          torch.Tensor

        """

        # check if first dimension of batch of inputs is the same
        if isinstance(speech, list):
            if len(speech) != len(text_inputs["input_ids"]):
                raise ValueError("Batch size of speech and text inputs should be the same")

        sampling_rate = self.feature_extractor.sampling_rate
        audio_inputs = self.feature_extractor(speech, sampling_rate=sampling_rate, return_tensors="pt", padding=True)
        audio_inputs = {key: audio_inputs[key].to(self._model.device) for key in audio_inputs}

        text_inputs = {key: text_inputs[key].to(self._model.device) for key in text_inputs}

        inputs = dict(
            input_values=audio_inputs["input_values"],
            audio_attention_mask=audio_inputs["attention_mask"],
            input_ids=text_inputs["input_ids"],
            token_type_ids=text_inputs["token_type_ids"],
            text_attention_mask=text_inputs["attention_mask"]
        )

        with torch.no_grad():
            logits = self._model(**inputs).logits

        # move inputs to cpu back
        for key in inputs:
            inputs[key] = inputs[key].cpu()

        del inputs

        if self._logistic_fct is torch.sigmoid:
            scores = self._logistic_fct(logits)
        elif self._logistic_fct is torch.softmax:
            scores = self._logistic_fct(logits, dim=1)
        else:
            raise ValueError('logistic_fct must be torch.sigmoid or torch.softmax')

        return scores

    def _recognize_one(self, path: str, text: str) -> RecognizerOutputOne:
        """Исполнение рантайма для одного файла

        Args:
          path: путь к файлу
          text: текст

        Returns:
          RecognizerOutputOne

        """
        speech = self.speech_file_to_array_fn(path)
        tokens = self.tokenize(text, self.tokenizer)
        scores = self._get_torch_scores(speech, tokens)

        scores = {k: v for k, v in zip(self.config.id2label.values(), scores[0].tolist())}

        return RecognizerOutputOne(**scores)

    def _recognize_many(self, voice_text_couple: List[Tuple[str, str]]) -> RecognizerOutputMany:
        """Прогнозируем несколько файлов

        Args:
          voice_text_couple: список путей к файлам

        Returns:
          словарь с выходами модели

        """
        paths = [path for path, _ in voice_text_couple]

        speeches = []
        tokens = []
        for path, text in voice_text_couple:
            speech = self.speech_file_to_array_fn(path)
            speeches.append(speech)
            tokens.append(text)

        tokens = self.tokenize(tokens, self.tokenizer)
        scores: torch.Tensor = self._get_torch_scores(speeches, tokens)
        results: RecognizerOutputMany = self._get_many_results(paths, scores)

        return results

    def recognize(
            self,
            paths_text_couple: Union[List[Tuple[str, str]], Tuple[str, str]],
            return_single_label: bool = False,
            top_n: Union[int, None] = None
    ) -> Union[RecognizerOutputOne, RecognizerOutputMany]:
        """Прогнозируем файлы

        Args:
          paths_text_couple: пары путь к файлу и текст (str, str) или список таких пар
          return_single_label: если True, то возвращаем только один лейбл
          top_n: количество лейблов, которые необходимо вернуть
        Returns:
          выход модели

        """
        if self._model is None:
            self._setup_variables()

        # check if config.problem_type is exists
        if self.config.problem_type is not None:
            if self.config.problem_type == 'multi_label_classification':
                return_single_label = False

        if isinstance(paths_text_couple, tuple):
            if return_single_label:
                return self._get_single_label(self._recognize_one(*paths_text_couple))

            elif top_n is not None:
                return self._get_top_n_labels(self._recognize_one(*paths_text_couple), top_n)

            return self._recognize_one(*paths_text_couple)

        elif isinstance(paths_text_couple, list):
            if return_single_label:
                return self._get_single_label(self._recognize_many(paths_text_couple))

            elif top_n is not None:
                return self._get_top_n_labels(self._recognize_many(paths_text_couple), top_n)

            return self._recognize_many(paths_text_couple)

        else:
            raise ValueError('paths_text_couple should be tuple or list')


class MultiModalRecognizer(VoiceTextRecognizer):
    def __init__(
            self,
            model: Model = None,
            s2t_model: Speech2Text = None,
            device: str = 'cpu',
            setup_on_init: bool = True
    ) -> None:
        super().__init__(model, device, setup_on_init)
        self.s2t_model = s2t_model

    def speech_to_text(self, speech: Union[torch.Tensor, numpy.ndarray, str]) -> str:
        """Преобразование речи в текст

        Args:
          speech: ndarray

        Returns:
          текст

        """
        return self.s2t_model.recognize(speech).text.strip()

    def _recognize_one(self, path: str, **kwargs) -> RecognizerOutputOne:
        """Исполнение рантайма для одного файла

        Args:
          path: путь к файлу
          text: текст

        Returns:
          RecognizerOutputOne

        """
        text = self.speech_to_text(path)

        return super()._recognize_one(path, text)

    def _recognize_many(self, paths: List[str]) -> RecognizerOutputMany:
        """Прогнозируем несколько файлов

        Args:
          voice_text_couple: список путей к файлам

        Returns:
          словарь с выходами модели

        """

        couples = []

        for path in paths:
            text = self.speech_to_text(path)
            couples.append((path, text))

        return super()._recognize_many(couples)

    def recognize(
            self,
            paths: Union[List[str], str],
            return_single_label: bool = False,
            top_n: Union[int, None] = None
    ) -> Union[RecognizerOutputOne, RecognizerOutputMany]:
        """Прогнозируем файлы

        Args:
          paths: путь к файлу или список таких путей
          return_single_label: если True, то возвращаем только один лейбл
          top_n: количество лейблов, которые нужно вернуть

        Returns:
          выход модели

        """
        if self._model is None:
            self._setup_variables()

        # check if config.problem_type is exists
        if self.config.problem_type is not None:
            if self.config.problem_type == 'multi_label_classification':
                return_single_label = False

        if isinstance(paths, str):
            if return_single_label:
                return self._get_single_label(self._recognize_one(paths))

            elif top_n is not None:
                return self._get_top_n_labels(self._recognize_one(paths), top_n)

            return self._recognize_one(paths)

        elif isinstance(paths, list):
            if return_single_label:
                return self._get_single_label(self._recognize_many(paths))

            elif top_n is not None:
                return self._get_top_n_labels(self._recognize_many(paths), top_n)

            return self._recognize_many(paths)

        else:
            raise ValueError('paths should be str or list')



"""Base model classes
"""
from aniemore.custom.modeling_classificators import BaseModelForVoiceBaseClassification
from transformers import HubertForSequenceClassification


class HubertForVoiceClassification(BaseModelForVoiceBaseClassification):
    """HubertForVoiceClassification is a model for voice classification task
     (e.g. speech command, voice activity detection, etc.)

    Args:
        config (HubertConfig): config
        num_labels (int): number of labels

    Attributes:
        config (HubertConfig): config
        num_labels (int): number of labels
        hubert (HubertForSequenceClassification): hubert model
    """

    def __init__(self, config, num_labels):
        super().__init__(config, num_labels)
        self.hubert = HubertForSequenceClassification(config)
        self.init_weights()



"""Base model classes
"""
from aniemore.custom.modeling_classificators import BaseModelForVoiceBaseClassification
from transformers import UniSpeechSatForSequenceClassification


class UniSpeechSatForVoiceClassification(BaseModelForVoiceBaseClassification):
    """UniSpeechSatForVoiceClassification is a model for voice classification task
     (e.g. speech command, voice activity detection, etc.)

    Args:
        config (UniSpeechSatConfig): config
        num_labels (int): number of labels

    Attributes:
        config (UniSpeechSatConfig): config
        num_labels (int): number of labels
        unispeech_sat (UniSpeechSatForSequenceClassification): unispeech_sat model
    """

    def __init__(self, config, num_labels):
        super().__init__(config, num_labels)
        self.unispeech_sat = UniSpeechSatForSequenceClassification(config)
        self.init_weights()



import numpy
import torch
import whisper
from typing import NamedTuple, List, Union
from functools import partial


class SpeechSegment(NamedTuple):
    """Структура для хранения результатов распознавания

    Attributes:
     id(int): локальный номер сегмента
     seek(int): смещение в секундах от начала аудио
     start(float): время начала сегмента в секундах
     end(float): время конца сегмента в секундах
     text(str): распознанный текст
     tokens(List[int]): список токенов
     temperature(float): whisper температура
     avg_logprob(float): средняя вероятность
     no_speech_prob(float): вероятность отсутствия речи
     compression_ratio(float): коэффициент сжатия
    """
    id: int
    seek: int
    start: float
    end: float
    text: str
    tokens: List[int]
    temperature: float
    avg_logprob: float
    no_speech_prob: float
    compression_ratio: float


class Speech2TextOutput(NamedTuple):
    """Структура для хранения результатов распознавания

    Attributes:
     text(str): распознанный текст
     language(str): язык
     segments(List[SpeechSegment]): список сегментов

    """
    text: str
    language: str
    segments: List[SpeechSegment]


class Speech2Text:
    def __init__(self, model_path: str):
        """Инициализация модели распознавания речи

        Args:
         model_path: путь к модели

        """
        self.model = whisper.load_model(model_path)

    def __call__(self, audio_path: str) -> Speech2TextOutput:
        """
        Распознать аудио

        Args:
         audio_path: путь к аудио

        Returns:
         результат распознавания

        >>> speech2text = Speech2Text('base')
        >>> speech2text('audio.wav')
        """
        return self.recognize(audio_path)

    def recognize(self, audio_path: Union[str, torch.Tensor, numpy.ndarray]) -> Speech2TextOutput:
        """Распознать аудио

        Args:
         audio_path: путь к аудио

        Returns:
          результат распознования

        Examples:
         >>> speech2text = Speech2Text('base')
         >>> speech2text.recognize('audio.wav')
        """
        result = self.model.transcribe(audio_path)
        result['segments'] = [SpeechSegment(**x) for x in result['segments']]
        return Speech2TextOutput(**result)


TinySpeech2Text = partial(Speech2Text, 'tiny')

BaseSpeech2Text = partial(Speech2Text, 'base')

SmallSpeech2Text = partial(Speech2Text, 'small')

MediumSpeech2Text = partial(Speech2Text, 'medium')

LargeSpeech2Text = partial(Speech2Text, 'large')




"""
Custom classes for aniemore
"""
import torch
import gc
from contextlib import contextmanager
from typing import ClassVar, ContextManager, Any, List, Union, NamedTuple, Dict, Type

import transformers
from transformers import (
    PretrainedConfig,
    AutoConfig,
    AutoTokenizer,
    AutoFeatureExtractor,
    BertForSequenceClassification,
    PreTrainedModel
)

from aniemore.custom.modeling_classificators import BaseMultiModalForSequenceBaseClassification
from aniemore.models import Model

RecognizerOutputOne: Type[Dict[str, float]] = dict


class RecognizerOutputTuple(NamedTuple):
    """Структура для хранения результатов распознавания"""
    key: str
    output: RecognizerOutputOne


RecognizerOutputMany: Type[Dict[str, RecognizerOutputOne]] = dict


class BaseRecognizer:
    # all examples of this class will be saved in this list
    CLASS_HANDLERS: ClassVar[List[Any]] = []

    def __init__(self, model: Model = None, device: str = 'cpu', setup_on_init: bool = True) -> None:
        """
        Инициализируем класс

        Args:
         model(aniemore.models.Model): название модели из `aniemore.custom.classes`
         device(str): 'cpu' or 'cuda' or 'cuda:<number>'
         setup_on_init(bool): если True, то сразу загружаем модель и токенайзер в память

        Examples:
         >>> from aniemore.models import HuggingFaceModel
         >>> from aniemore.recognizers.text import TextRecognizer
         >>> tr = TextRecognizer(HuggingFaceModel.Text.Bert_Tiny, device='cuda:0')
         >>> tr.recognize('Как же я люблю природу, она прекрасна!!)))')
        """
        self._model: Any = None
        self._device: Union[str, None] = None
        self.config: Type[PretrainedConfig] = None
        self.model_cls: Type[PreTrainedModel] = None
        self.model_url: Union[str, None] = None
        self._logistic_fct: Any = None

        self.model = model
        self.device = device
        self._add_to_class_handlers()
        if setup_on_init:
            self._setup_variables()

    def _setup_variables(self) -> None:
        """Загружаем модель и экстрактор признаков в память
        """
        # check if model_cls is child of BaseMultiModalForSequenceClassification
        if issubclass(self.model_cls, BaseMultiModalForSequenceBaseClassification):
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_url)
            self.feature_extractor = AutoFeatureExtractor.from_pretrained(self.model_url)
        elif self.model_cls is BertForSequenceClassification:
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_url)
        else:
            self.feature_extractor = AutoFeatureExtractor.from_pretrained(self.model_url)

        try:
            self.config = AutoConfig.from_pretrained(self.model_url)
            self._model = self.model_cls.from_pretrained(self.model_url, config=self.config)
        except (RuntimeError, ValueError) as exc:
            self.config = AutoConfig.from_pretrained(self.model_url, trust_remote_code=True)
            self._model = self.model_cls.from_pretrained(
                self.model_url, trust_remote_code=True, config=self.config
            )
        finally:
            self._model = self._model.to(self.device)

        # check problem type and set logistic function (sigmoid or softmax, else softmax)
        if self.config.problem_type == 'single_label_classification':
            self._logistic_fct = torch.softmax
        elif self.config.problem_type == 'multi_label_classification':
            self._logistic_fct = torch.sigmoid
        else:
            self._logistic_fct = torch.softmax

    # add example to the list
    def _add_to_class_handlers(self):
        """Добавляем экземпляр класса в список всех экземпляров этого класса"""
        self.CLASS_HANDLERS.append(self)

    # remove example from the list
    def _remove_from_class_handlers(self):
        """Удаляем экземпляр класса из списка всех экземпляров этого класса"""
        if self in self.CLASS_HANDLERS:
            self.CLASS_HANDLERS.remove(self)

    # get all examples that is not an existing example of this class
    def _get_class_handlers(self):
        """Получаем список всех экземпляров этого класса, кроме текущего"""
        return [handler for handler in self.CLASS_HANDLERS if handler is not self]

    @property
    def device(self) -> str:
        """Возвращаем устройство, на котором будет работать модель"""
        return self._device

    @device.setter
    def device(self, value) -> None:
        """Устанавливаем устройство, на котором будет работать модель

        Args:
          value: возможные значения: 'cpu', 'cuda', 'cuda:<number>'

        Returns:
          None or raises ValueError

        """
        if value != 'cpu':
            if not self.validate_device(value):
                raise ValueError(f"Device must be 'cpu' or 'cuda', 'cuda:<number>' or 'mps', not {value}")
        self._device = value

        # set model to the given device
        if self._model is not None:
            self._model.to(self.device)

            if value != 'cpu' and torch.cuda.is_available():
                torch.cuda.empty_cache()

    @classmethod
    def validate_device(cls, value) -> bool:
        """Валидатор для устройства, на котором будет работать модель

        Args:
          value: возможные значения: 'cpu', 'cuda', 'cuda:<number>'

        Returns:
          True or False

        """
        try:
            torch.device(value)
            return True
        except RuntimeError:  # torch device error
            return False
        # if value != 'cpu':
        #     if re.match(r'^(cuda)(:\d+)?$', value) is None:  # https://regex101.com/r/SGEiYz/2
        #         return False
        # return True

    @property
    def model(self) -> Model:
        """Возвращаем текущую модель, которая будет распозновать данные

        Returns:
            модель, которая загружена в текущий момент
        """
        return Model(model_cls=self.model_cls, model_url=self.model_url)

    @model.setter
    def model(self, model: Model) -> None:
        """Устанавливаем модель, которая будет распозновать данные

        Args:
          model: валидная модель (тип модели смотрите в `aniemore.config.Model`)

        Returns:
          None

        """
        if self.validate_model(model):
            self.model_cls, self.model_url = model
            self._model = self.model_cls.from_pretrained(self.model_url)
        else:
            raise ValueError('Not a valid model provided: %s', model)

    @classmethod
    def validate_model(cls, model: Model) -> bool:
        """
        Валидатор для загружаемой модели
        Args:
          model: модель из `models.py`

        Returns:
         `True` если прошел валидацию, `False` если не прошёл
        """
        return all([
            isinstance(model, Model),
            isinstance(model.model_url, str),
            issubclass(model.model_cls, transformers.PreTrainedModel),
        ])

    # create a context manager that allows this proof of work:
    @contextmanager
    def on_device(
            self,
            device: Union[str, torch.device],
            clear_same_device_cache: bool = True,
            clear_cache_after: bool = True) -> ContextManager:
        """Context manager that allows you to switch the model to the given device

        Args:
          device: cpu' or 'cuda' or 'cuda:<number>'
          clear_same_device_cache: clear cuda cache after switching to the given device
          clear_cache_after: clear cuda cache after switching to the original device

        Returns:
          None

        Examples:
         >>> with model.on_device('cuda'):
         >>>     # do something
        """
        try:
            # get other examples of this class and switch them to cpu device
            for handler in self._get_class_handlers():
                # check if the device is already the same
                if handler.device == device and clear_same_device_cache:
                    # move to cpu
                    handler._model = handler._model.to('cpu')

            # switch this example to the given device
            self._model = self._model.to(device)

            # clear cuda cache
            if clear_same_device_cache and torch.cuda.is_available():
                torch.cuda.empty_cache()

            yield
        finally:
            # switch this example to original device
            self._model = self._model.to(self._device)

            # clear cuda cache
            if clear_cache_after and torch.cuda.is_available():
                torch.cuda.empty_cache()

            # get other examples of this class and switch them to their original device
            for handler in self._get_class_handlers():
                handler._model = handler._model.to(handler.device)

            # do garbage collection
            if clear_cache_after:
                gc.collect()

    @contextmanager
    def with_model(
            self,
            model: Model,
            device: Union[str, torch.device],
            clear_cache_after: bool = True) -> ContextManager:
        """Context manager that allows you to switch the model to the given model

        Args:
          model: model
          device: cpu' or 'cuda' or 'cuda:<number>'
          clear_cache_after: clear cuda cache after switching to the original device

        Returns:
          None

        Examples:
         >>> with vr.with_model(new_model, 'cuda') as new_model:
         >>>     # do something
        """
        new_handler = None

        try:
            # create new example of this class with the given model
            new_handler = self.__class__(model=model, device=device, setup_on_init=True)
            yield new_handler
        finally:
            # delete new example
            del new_handler

            # clear cuda cache
            if clear_cache_after and torch.cuda.is_available():
                torch.cuda.empty_cache()

            # do garbage collection
            if clear_cache_after:
                gc.collect()

    def _get_torch_scores(self, *args, **kwargs) -> torch.Tensor:
        """[PROTECTED METHOD] Получаем тензор с предсказаниями модели

        Args:
          args: аргументы
          kwargs: аргументы

        Returns:
          тензор с предсказаниями модели

        """
        raise NotImplementedError

    def recognize(self, *args, **kwargs) -> Union[RecognizerOutputOne, RecognizerOutputMany]:
        """Получаем предсказания модели

        Args:
          args: аргументы
          kwargs: аргументы

        Returns:
          предсказания модели

        """
        raise NotImplementedError

    def _recognize_one(self, *args, **kwargs) -> RecognizerOutputOne:
        """[PROTECTED METHOD] Получаем предсказания модели для одного объекта

        Args:
          args: аргументы
          kwargs: аргументы

        Returns:
          предсказания модели

        """
        raise NotImplementedError

    def _recognize_many(self, *args, **kwargs) -> RecognizerOutputMany:
        """[PROTECTED METHOD] Получаем предсказания модели для нескольких объектов

        Args:
          args: аргументы
          kwargs: аргументы

        Returns:
          предсказания модели

        """
        raise NotImplementedError

    def _get_many_results(self, items: List[str], scores: torch.Tensor) -> RecognizerOutputMany:
        """[PROTECTED METHOD] Принимает на вход исследуемые объекты и результат от модели,
        и возвращает результат

        Args:
          items: список исследуемых объектов
          scores: выход от функции _get_torch_scores

        Returns:
          dict` с результатами по каждому объекту

        """
        result = []
        for path_, score in zip(items, scores):
            score = {k: v for k, v in zip(self.config.id2label.values(), score.tolist())}
            result.append(RecognizerOutputTuple(path_, RecognizerOutputOne(**score)))
        return RecognizerOutputMany(tuple(result))

    @classmethod
    def _get_single_label(
            cls,
            output: Union[RecognizerOutputOne, RecognizerOutputMany]
    ) -> Union[str, dict]:
        """[PROTECTED CLASS METHOD] Получаем метку из предсказаний модели

        Args:
          output: предсказания модели

        Returns:
          метка

        """
        # check if output is dict of [str: float]
        if isinstance(output, dict) and all(isinstance(x, float) for x in output.values()):
            # max score in dict
            return max(output, key=output.get)

        # check if output is dict of [str: dict[str: float]]
        if isinstance(output, dict) and all(isinstance(x, dict) for x in output.values()):
            # max score in list
            return {x: max(output[x], key=output[x].get) for x in output.keys()}

    @classmethod
    def _get_top_n_labels(
            cls,
            output: Union[RecognizerOutputOne, RecognizerOutputMany],
            n: int
    ) -> List[str] | dict:
        """[PROTECTED CLASS METHOD] Получаем метку из предсказаний модели

        Args:
            output: предсказания модели
            n: количество меток, которые нужно вернуть

        Returns:
            метка

        """
        # check if output is dict of [str: float]
        if isinstance(output, dict) and all(isinstance(x, float) for x in output.values()):
            # max score in dict
            return sorted(output, key=output.get, reverse=True)[:n]

        # check if output is dict of [str: dict[str: float]]
        if isinstance(output, dict) and all(isinstance(x, dict) for x in output.values()):
            # max score in list
            return {x: sorted(output[x], key=output[x].get, reverse=True)[:n] for x in output.keys()}

    def __del__(self):
        """
        Деструктор

        :return: None
        """
        self._remove_from_class_handlers()
        if torch.cuda.is_available():
            torch.cuda.empty_cache()



"""Base model classes
"""
from aniemore.custom.modeling_classificators import (
    BaseModelForVoiceBaseClassification,
    BaseMultiModalForSequenceBaseClassification
)

import torch
from transformers import (
    Wav2Vec2ForSequenceClassification,
    BertConfig,
    BertModel,
    Wav2Vec2Config,
    Wav2Vec2Model
)

from transformers.models.wav2vec2.modeling_wav2vec2 import (
    Wav2Vec2Encoder,
    Wav2Vec2EncoderStableLayerNorm,
    Wav2Vec2FeatureEncoder
)

from transformers.models.bert.modeling_bert import BertEncoder


class Wav2Vec2ForVoiceClassification(BaseModelForVoiceBaseClassification):
    """Wav2Vec2ForVoiceClassification is a model for voice classification task
     (e.g. speech command, voice activity detection, etc.)

    Args:
        config (Wav2Vec2Config): config
        num_labels (int): number of labels

    Attributes:
        config (Wav2Vec2Config): config
        num_labels (int): number of labels
        wav2vec2 (Wav2Vec2ForSequenceClassification): wav2vec2 model
    """

    def __init__(self, config, num_labels):
        super().__init__(config, num_labels)
        self.wav2vec2 = Wav2Vec2ForSequenceClassification(config)
        self.init_weights()


class Wav2Vec2BertForSequenceClassification(BaseMultiModalForSequenceBaseClassification):
    """Wav2Vec2BertForSequenceClassification is a model for sequence classification task
     (e.g. sentiment analysis, text classification, etc.)

    Args:
        config (Wav2Vec2BertConfig): config

    Attributes:
        config (Wav2Vec2BertConfig): config
        audio_config (Wav2Vec2Config): wav2vec2 config
        text_config (BertConfig): bert config
        audio_model (Wav2Vec2Model): wav2vec2 model
        text_model (BertModel): bert model
        classifier (torch.nn.Linear): classifier
    """

    def __init__(self, config, finetune=False):
        super().__init__(config)
        self.supports_gradient_checkpointing = getattr(config, "gradient_checkpointing", True)

        self.audio_config = Wav2Vec2Config.from_dict(self.config.Wav2Vec2Model)
        self.text_config = BertConfig.from_dict(self.config.BertModel)

        if not finetune:
            self.audio_model = Wav2Vec2Model(self.audio_config)
            self.text_model = BertModel(self.text_config)

        else:
            self.audio_model = Wav2Vec2Model.from_pretrained(self.audio_config._name_or_path, config=self.audio_config)
            self.text_model = BertModel.from_pretrained(self.text_config._name_or_path, config=self.text_config)

        self.classifier = torch.nn.Linear(
            self.audio_config.hidden_size + self.text_config.hidden_size, self.num_labels
        )
        self.init_weights()

    @staticmethod
    def _set_gradient_checkpointing(module, value=False):
        if isinstance(module, (Wav2Vec2Encoder, Wav2Vec2EncoderStableLayerNorm, Wav2Vec2FeatureEncoder, BertEncoder)):
            module.gradient_checkpointing = value
